/*
Copyright (C) Arturo Torres Sánchez 2013

This file is part of Trino.

Trino is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Trino is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Trino.  If not, see <http://www.gnu.org/licenses/>.
*/

using Gee;
using Gtk;
using AppIndicator;
using WebKit;

namespace Trino {
    public class Main : Granite.Application {
        construct {
            application_id = "org.pantheon.trino";
            exec_name = "trino";
            app_launcher = "trino.desktop";
            
            program_name = "Trino";
            build_version = "0.1";
            app_icon = "twitter";
            
            about_comments = "Twitter client";
            app_years = "2013";
            main_url = "http://launchpad.net/trino";
            
            about_authors =
                {"Arturo Torres Sánchez <r2d2.art2005@gmail.com>", null};
            about_documenters =
                {"Arturo Torres Sánchez <r2d2.art2005@gmail.com>", null};
            about_artists =
                {"Arturo Torres Sánchez <r2d2.art2005@gmail.com>", null};
            about_translators = "Launchpad Translators";
            about_license_type = License.GPL_3_0;
            
            help_url = "http://answers.launchpad.net/trino";
            translate_url = "http://translations.launchpad.net/trino";
            bug_url = "http://bugs.launchpad.net/trino";
        }
        
        private GLib.Settings settings;
        private Gee.List<WebView> views;
        private Gee.List<Granite.Widgets.SourceList.Item> items;
        private Gee.List<Twitter> accounts;
        private Map<string, string> cached_tweets;
        private Gtk.Notebook page_switcher;
        private Unity.LauncherEntry launcher;
        
        public Main() {
            settings = new GLib.Settings ("org.pantheon.trino");
            views = new Gee.ArrayList<WebView> ();
            accounts = new Gee.ArrayList<Twitter> ();
            items = new Gee.ArrayList<Granite.Widgets.SourceList.Item> ();
            cached_tweets = new HashMap<string, string> ();
            launcher = Unity.LauncherEntry.get_for_desktop_id ("trino.desktop");
        }
        
        private void on_tweet_button_clicked () {
            create_new_tweet (null, null);
        }
        private void update_launcher () {
            int unreadAll = 0;
            foreach (Twitter twitter in accounts) {
                var unread = twitter.get_unread_count ();
                unreadAll += unread;
                var item = items.get ((accounts.index_of (twitter) * 2) + 1); //Mentions item
                if (unread == 0) {
                    item.badge = "";
                } else {
                    item.badge = unread.to_string ();
                }
            }
            launcher.count = unreadAll;
            launcher.count_visible = (unreadAll != 0);
        }
        private void create_new_tweet (string? id, string? screen_name) {
            var is_reply = (id != null);
            string image = null;
            var window = new Granite.Widgets.LightWindow ();
            Twitter twitter = accounts.get (page_switcher.get_current_page () / 2);
            window.window_position = WindowPosition.CENTER_ALWAYS;
            if (is_reply) {
                window.title = "Reply to " + screen_name;
            } else {
                window.title = "New Tweet";
            }
            window.set_size_request (448,150);
            window.set_default_size (448,150);
            window.border_width = 8;
            
            var webview = new WebView ();
            if (is_reply) {
                webview.set_size_request (400, 48);
                webview.navigation_policy_decision_requested.connect ((frame, request, action, decision) => {
                    if (!request.uri.has_suffix ("single.html")) {
                        decision.ignore ();
                        return true;
                    }
                    return false;
                });
                File file = File.new_for_path ("/usr/local/share/trino/html/single.html");
                webview.load_uri (file.get_uri ());
                string tweet = cached_tweets.get (id);
                webview.document_load_finished.connect ((frame) => {
                    webview.execute_script ("appendMessage('%s')".printf (tweet));
                    var dom = webview.get_dom_document ();
                    var chat = dom.get_element_by_id ("Chat");
                    var height = chat.get_offset_height ();
                    webview.set_size_request (400, (int) height);
                });
            }
            
            var textview = new TextView ();
            var tweet_button = new Gtk.Button.with_label("Tweet");
            var attach_button = new Gtk.Button.with_label("Add image...");
            
            textview.wrap_mode = WrapMode.WORD_CHAR;
            textview.set_size_request (400, 48);
            textview.key_press_event.connect ((event) => {
                if ((event.keyval == Gdk.Key.Return) &&
                    ((event.state & Gdk.ModifierType.SHIFT_MASK) == 0)) {
                    tweet_button.clicked ();
                    return true;
                }
                return false;
            });
            
            var count = new Gtk.Label ("140");
            var url_label = new Gtk.Label ("");
            url_label.margin_left = 20;
            
            var buffer = textview.buffer;
            buffer.modified_changed.connect (() => {
                var url_count = Trino.Parser.get_url_count (buffer.text, twitter.short_url_length, twitter.short_url_length_https);
                if (url_count != buffer.text.char_count ()) {
                    url_label.set_text ("Links will be shortened");
                } else {
                    url_label.set_text ("");
                }
                if (image != null) {
                    count.set_text ((140 - twitter.characters_reserved_per_media - url_count).to_string());
                } else {
                    count.set_text ((140 - url_count).to_string());
                }
                buffer.set_modified (false);
            });
            if (screen_name != null) {
                buffer.text = "@" + screen_name + " ";
            }
            attach_button.clicked.connect (() => {
                var chooser = new Gtk.FileChooserDialog (
                    "Select image to upload", window, Gtk.FileChooserAction.OPEN,
                    Gtk.Stock.CANCEL,
                    Gtk.ResponseType.CANCEL,
                    Gtk.Stock.OPEN,
                    Gtk.ResponseType.ACCEPT);
                chooser.select_multiple = false;
                
                var filter = new Gtk.FileFilter ();
                filter.set_filter_name ("Images (PNG, JPEG, GIF)");
                filter.add_mime_type ("image/png");
                filter.add_mime_type ("image/jpeg");
                filter.add_mime_type ("image/gif");
                chooser.add_filter (filter);
                
                var response = chooser.run ();
                if (response == Gtk.ResponseType.ACCEPT) {
                    image = chooser.get_filename ();
                } else if (response == Gtk.ResponseType.CANCEL) {
                    image = null;
                }
                buffer.set_modified (true);
                chooser.destroy ();
            });
            tweet_button.clicked.connect (() => {
                uint response;
                if (image != null) {
                    response = twitter.send_image (buffer.text, id, image);
                } else {
                    response = twitter.send_tweet (buffer.text, id);
                }
                if (response == 200) {
                    window.destroy ();
                }
            });

            var vbox = new Gtk.Box (Gtk.Orientation.VERTICAL, 0);
            if (is_reply)
                vbox.pack_start (webview, true, true, 0);
            vbox.pack_start (textview, true, true, 0);
            var hbox = new Gtk.Box (Gtk.Orientation.HORIZONTAL, 0);
            hbox.pack_start (count, false, false, 0);
            hbox.pack_start (url_label, false, false, 0);
            hbox.pack_end (tweet_button, false, false, 0);
            hbox.pack_end (attach_button, false, false, 0);
            vbox.pack_start (hbox, false, false, 0);
            
            window.add (vbox);
            window.show_all ();
            textview.grab_focus ();
        }
        private void create_retweet (string? id, string? screen_name, WebFrame frame) {
            var window = new Granite.Widgets.LightWindow ();
            Twitter twitter = accounts.get (page_switcher.get_current_page () / 2);
            window.window_position = WindowPosition.CENTER_ALWAYS;
            if (screen_name == null) {
                var user_id = twitter.user_id;
                window.title = user_id + " New Tweet";
            } else {
                window.title = "Retweet " + screen_name;
            }
            window.set_size_request (448,150);
            window.set_default_size (448,150);
            window.border_width = 8;
            
            var view = new WebView ();
            view.set_size_request (400, 48);
            view.navigation_policy_decision_requested.connect ((frame, request, action, decision) => {
                if (!request.uri.has_suffix ("single.html")) {
                    decision.ignore ();
                    return true;
                }
                return false;
            });
            File file = File.new_for_path ("/usr/local/share/trino/html/single.html");
            view.load_uri (file.get_uri ());
            string tweet = cached_tweets.get (id);
            view.document_load_finished.connect ((frame) => {
                view.execute_script ("appendMessage('%s')".printf (tweet));
                var dom = view.get_dom_document ();
                var chat = dom.get_element_by_id ("Chat");
                var height = chat.get_offset_height ();
                view.set_size_request (400, (int) height);
            });
            
            var tweet_button = new Gtk.Button.with_label("Retweet");
            tweet_button.clicked.connect (() => {
                var response = twitter.retweet (id);
                if (response == 200) {
                    window.destroy ();
                    frame.get_web_view ().execute_script ("addRetweet('%s')".printf (id));
                }
            });

            var vbox = new Gtk.Box (Gtk.Orientation.VERTICAL, 0);
            vbox.pack_start (view, true, true, 0);
            var hbox = new Gtk.Box (Gtk.Orientation.HORIZONTAL, 0);
            hbox.pack_end (tweet_button, false, false, 0);
            vbox.pack_start (hbox, false, false, 0);
            
            window.add (vbox);
            window.show_all ();
        }
        private void create_favorite (string? id, string? screen_name, WebFrame frame) {
            Twitter twitter = accounts.get (page_switcher.get_current_page () / 2);
            var response = twitter.favorite (id);
            if (response == 200) {
                frame.get_web_view ().execute_script ("addFavorite('%s')".printf (id));
            }
        }
        private void delete_tweet (string? id, WebFrame frame) {
            Twitter twitter = accounts.get (page_switcher.get_current_page () / 2);
            var response = twitter.delete_tweet (id);
            if (response == 200) {
                frame.get_web_view ().execute_script ("deleteMessage('%s')".printf (id));
            }
        }
        private void create_unread (string? id) {
            var index = (int) page_switcher.get_current_page () / 2;
            var home_view = views.get (index * 2);
            var mentions_view = views.get (index * 2 + 1);
            Twitter twitter = accounts.get (index);
            if (twitter.is_unread (id)) {
                twitter.set_unread (id, false);
                home_view.execute_script ("removeUnread('%s')".printf (id));
                mentions_view.execute_script ("removeUnread('%s')".printf (id));
            } else {
                twitter.set_unread (id, true);
                home_view.execute_script ("addUnread('%s')".printf (id));
                mentions_view.execute_script ("addUnread('%s')".printf (id));
            }
            update_launcher ();
        }
        private void show_conversation (string? id) {
            Twitter twitter = accounts.get (page_switcher.get_current_page () / 2);
            var conv = new Conversation ();
            conv.view.document_load_finished.connect ((frame) => {
                string tweet_id = id;
                do {
                    stdout.printf (id + "\n");
                    var tweet = twitter.get_tweet (tweet_id);
                    if (tweet == null) {
                        //stdout.printf ("null\n");
                    }
                    process_tweets (tweet, frame.get_web_view (), twitter);
                    if (tweet.has_member("in_reply_to_status_id_str")) {
                        tweet_id = tweet.get_string_member ("in_reply_to_status_id_str");
                    } else {
                        //stdout.printf ("it hasnt\n");
                        tweet_id = null;
                    }
                } while (tweet_id != null);
            });
            conv.show_all ();
        }
        
        private Gtk.Widget create_welcome () {
            /*var welcome = new Granite.Widgets.Welcome (
                "You have no tweets",
                "Configure Trino to start tweeting");
            welcome.expand = true;
            welcome.append("twitter","Add account","Link a Twitter account");
            welcome.activated.connect((index) => {
            });*/
            //var window = new Gtk.Window ();
            //window.title = "Authentication";
            //window.set_default_size (800,600);
            var view = new WebView();
            var scroll = new ScrolledWindow (null, null);
            scroll.add (view);
            scroll.expand = true;
            //window.add (scroll);
            //window.show_all ();
            Twitter.get_access_token (view);
            return scroll;
        }
        
        private bool handle_links (WebFrame frame, NetworkRequest request, WebNavigationAction action, WebPolicyDecision decision) {
            var uri = request.get_uri();
            if (uri.has_prefix ("http://t.co/") || uri.has_prefix ("https://t.co/")) {
                try {
                    AppInfo.launch_default_for_uri (uri, null);
                } catch (GLib.Error e) {
                    stderr.printf ("Couldn't open URI");
                }
                decision.ignore ();
                return true;
            } else if (uri.has_prefix ("trino://delete")) {
                var response = uri.replace ("trino://delete?", "");
                var map = Trino.Parser.parse_response (response);
                var id = map.get ("id");
                delete_tweet (id, frame);
                decision.ignore ();
                return true;
            } else if (uri.has_prefix ("trino://reply")) {
                var response = uri.replace ("trino://reply?", "");
                var map = Trino.Parser.parse_response (response);
                var id = map.get ("id");
                var name = map.get ("screen_name");
                create_new_tweet (id, name);
                decision.ignore ();
                return true;
            } else if (uri.has_prefix ("trino://retweet")) {
                var response = uri.replace ("trino://retweet?", "");
                var map = Trino.Parser.parse_response (response);
                var id = map.get ("id");
                var name = map.get ("screen_name");
                create_retweet (id, name, frame);
                decision.ignore ();
                return true;
            } else if (uri.has_prefix ("trino://favorite")) {
                var response = uri.replace ("trino://favorite?", "");
                var map = Trino.Parser.parse_response (response);
                var id = map.get ("id");
                var name = map.get ("screen_name");
                create_favorite (id, name, frame);
                decision.ignore ();
                return true;
            } else if (uri.has_prefix ("trino://unread")) {
                var response = uri.replace ("trino://unread?", "");
                var map = Trino.Parser.parse_response (response);
                var id = map.get ("id");
                create_unread (id);
                decision.ignore ();
                return true;
            } else if (uri.has_prefix ("trino://conversation")) {
                var response = uri.replace ("trino://conversation?", "");
                var map = Trino.Parser.parse_response (response);
                var id = map.get ("id");
                show_conversation (id);
                decision.ignore ();
                return true;
            }
            return false;
        }
        
        private Granite.Widgets.ThinPaned create_paned () {
            var sidebar = new Granite.Widgets.SourceList ();
            
            page_switcher = new Gtk.Notebook ();
            page_switcher.show_tabs = false;
            page_switcher.show_border = false;
            page_switcher.expand = true;
            
            var users = settings.get_strv ("users");
            foreach (string user_id in users) {
                var twitter = new Twitter (user_id);
                accounts.add (twitter);
                
                var screen_name = twitter.screen_name;
                var cat = new Granite.Widgets.SourceList.ExpandableItem (screen_name);
                
                var home_view = new WebView ();
                //Developer stuff (should be removed)
                home_view.settings.enable_developer_extras = true;
                home_view.web_inspector.inspect_web_view.connect ((p0) => {
                    var w = new Window ();
                    w.window_position = WindowPosition.CENTER;
                    w.title = "Inspector";
                    w.set_default_size (800,600);
                    WebView view = new WebView ();
                    unowned WebView view2 = view;
                    w.add (view2);
                    w.show_all ();
                    return view2;
                });
                home_view.navigation_policy_decision_requested.connect (handle_links);
                File file = File.new_for_path ("/usr/local/share/trino/html/timeline.html");
                home_view.load_uri (file.get_uri ());
                home_view.document_load_finished.connect ((frame) => {
                    twitter.home_timeline(process_tweets, home_view);
                });
                var scroll1 = new Gtk.ScrolledWindow (null, null);
                scroll1.add (home_view);
                views.add (home_view);
                page_switcher.append_page (scroll1, null);
                var item1 = new Granite.Widgets.SourceList.Item ("Home");
                items.add (item1);
                cat.add (item1);
                
                var mentions_view = new WebView ();
                //Developer stuff (should be removed)
                mentions_view.settings.enable_developer_extras = true;
                mentions_view.web_inspector.inspect_web_view.connect ((p0) => {
                    var w = new Window ();
                    w.window_position = WindowPosition.CENTER;
                    w.title = "Inspector";
                    w.set_default_size (800,600);
                    WebView view = new WebView ();
                    unowned WebView view2 = view;
                    w.add (view2);
                    w.show_all ();
                    return view2;
                });
                mentions_view.navigation_policy_decision_requested.connect (handle_links);
                mentions_view.load_uri (file.get_uri ());
                mentions_view.document_load_finished.connect ((frame) => {
                    twitter.mentions_timeline(process_tweets, mentions_view);
                });
                var scroll2 = new Gtk.ScrolledWindow (null, null);
                scroll2.add (mentions_view);
                views.add (mentions_view);
                page_switcher.append_page (scroll2);
                var item2 = new Granite.Widgets.SourceList.Item ("Mentions");
                items.add (item2);
                cat.add (item2);
                
                //var item2 = new Granite.Widgets.SourceList.Item ("Mentions");
                //items.add (item2);
                //item1.activatable = new ThemedIcon.with_default_fallbacks ("mail-unread");
                //item2.badge = "5";
                //cat.add (item2);
                sidebar.root.add (cat);
            }
            sidebar.item_selected.connect ((item) => {
                page_switcher.set_current_page (items.index_of (item));
            });
            sidebar.root.expand_all ();
            
            var sidebar_paned = new Granite.Widgets.ThinPaned ();
            sidebar_paned.pack1 (sidebar, false, false);
            sidebar_paned.pack2 (page_switcher, true, true);
            sidebar_paned.expand = true;
            sidebar_paned.position = 200;
            return sidebar_paned;
        }
        
        private Window create_main_window () {
            var window = new Window();
            window.title = "Trino";
            window.window_position = WindowPosition.CENTER;
            window.set_default_size(800, 600);
            //window.set_application(this);
            add_window (window);
            
            var main_layout = new Grid ();
            main_layout.expand = true;
            main_layout.orientation = Orientation.VERTICAL;
            
            var main_toolbar = new Toolbar ();
            main_toolbar.get_style_context ()
                .add_class (STYLE_CLASS_PRIMARY_TOOLBAR);
            main_toolbar.hexpand = true;
            main_toolbar.vexpand = false;
            
            var tweet = new ToolButton (null, "Tweet");
            tweet.set_icon_name ("im-message-new");
            tweet.set_tooltip_text ("Tweet");
            tweet.clicked.connect (on_tweet_button_clicked);
            
            var separator = new SeparatorToolItem ();
            separator.draw = false;
            separator.set_expand (true);
            
            var main_menu = create_appmenu (new Gtk.Menu ());
            
            main_toolbar.insert (tweet, -1);
            main_toolbar.insert (separator, -1);
            main_toolbar.insert (main_menu, -1);
            
            main_layout.add (main_toolbar);
            bool first_run = settings.get_boolean ("first-run");
            if (first_run) {
                main_layout.add (create_welcome ());
            } else {
                main_layout.add (create_paned ());
            }
            
            window.add (main_layout);
            window.show_all ();
            window.set_focus (null);
            
            ////////
            
            var indicator = new Indicator(window.title, "indicator-messages",
                                          IndicatorCategory.APPLICATION_STATUS);

            indicator.set_status(IndicatorStatus.ACTIVE);
            indicator.set_attention_icon("indicator-messages-new");

            var menu = new Gtk.Menu();

            var item = new Gtk.MenuItem.with_label("Foo");
            item.activate.connect(() => {
                    indicator.set_status(IndicatorStatus.ATTENTION);
            });
            item.show();
            menu.append(item);

            item = new Gtk.MenuItem.with_label("Bar");
            item.show();
            item.activate.connect(() => {
                    indicator.set_status(IndicatorStatus.ATTENTION);
            });
            menu.append(item);

            indicator.set_menu(menu);
            
            return window;
        }
        
        public override void activate() {
            Notify.init("trino");
            unowned GLib.List<weak Window> windows = this.get_windows();
            if (windows != null) {
                var window = windows.first().data;
                window.present();
            } else {
                create_main_window ();
            }
        }
        private void map_entities (Json.Array array, uint index_, Json.Node node, Map<int, Json.Object> map) {
            if (!node.is_null()) {
                var obj = node.get_object ();
                var index = (int) obj.get_array_member ("indices").get_int_element (0);
                map.set (index, obj);
            }
        }
        private void process_media (Json.Object obj, ArrayList<string> media_list) {
			var media_builder = new StringBuilder ();
			var media_url = "";
			if (obj.has_member ("media_url_https")) {
				media_url = obj.get_string_member ("media_url_https") + ":thumb";
			} else if (obj.has_member ("url")) {
				var expanded_url = obj.get_string_member ("expanded_url");
				/* I was getting thumbnails from each site, but I found embed.ly which has support for 250 sites.
				if (expanded_url.has_prefix ("http://twitpic.com/")) {
				    media_url = expanded_url.replace ("http://twitpic.com/", "http://twitpic.com/show/thumb/");
				} else if (expanded_url.has_prefix ("http://imgur.com/")) {
				    media_url = expanded_url.replace ("http://imgur.com/", "http://i.imgur.com/") + "t.jpg";
				} else if (expanded_url.has_prefix ("http://i.imgur.com/")) {
				    media_url = expanded_url.replace (".jpg", "t.jpg");
				} else if (expanded_url.has_prefix ("http://instagram.com/p/") || expanded_url.has_prefix ("http://instagr.am/p/")) {
				    media_url = expanded_url + "media/?size=t";
				} else if (expanded_url.has_prefix ("http://ow.ly/i/")) {
				    media_url = expanded_url.replace ("http://ow.ly/i/", "http://static.ow.ly/photos/thumb/") + ".jpg";
				} else if (expanded_url.has_prefix ("http://yfrog.com/")) {
				    media_url = expanded_url + ":small";
				} else if (expanded_url.has_prefix ("http://youtu.be/")) {
				    media_url = expanded_url.replace ("http://youtu.be/","http://img.youtube.com/vi/") + "/default.jpg";
			    } else if (expanded_url.has_prefix ("http://youtube.com/") || expanded_url.has_prefix ("http://www.youtube.com/")) {
				    var response = Trino.Parser.parse_response (expanded_url);
				    var v = response.get ("v");
				    media_url = "http://img.youtube.com/vi/" + v + "/default.jpg";
				}*/
				var encoded_url = Soup.URI.encode (expanded_url, " !\"#$%&'(),:;<=>?[\\]^`{|}~");
				var message = new Soup.Message ("GET", "http://api.embed.ly/1/oembed?key=312a3f7894224247b2f7b776ccc5d4ce&url=" + encoded_url);
				stdout.printf ("http://api.embed.ly/1/oembed?key=312a3f7894224247b2f7b776ccc5d4ce&url=" + encoded_url + "\n");
				var session = new Soup.SessionAsync ();
				session.send_message(message);
				var response = (string) message.response_body.data;
				if (message.status_code == 200) {
					var parser = new Json.Parser ();
                    try {
				        stdout.printf (response + "\n");
                        parser.load_from_data (response);
                        var oembed = parser.get_root ().get_object ();
                        var type = oembed.get_string_member ("type");
                        if (type == "photo" || type == "video") {
                            if (oembed.has_member ("thumbnail_url")) {
                                media_url = oembed.get_string_member ("thumbnail_url");
                            } else {
                                media_url = oembed.get_string_member ("url");
                            }
                        }
                    } catch (GLib.Error e) {
                    }
				} else {
				    stdout.printf ("%d %s\n", (int) message.status_code, message.reason_phrase);
				}
			}
			if (media_url != "") {
			    media_builder.erase ();
			    media_builder.append ("<br /><a href=\"");
			    media_builder.append (obj.get_string_member ("url"));
			    media_builder.append ("\"><img class=\"media\" src=\"");
			    media_builder.append (media_url);
			    media_builder.append ("\" title=\"");
			    media_builder.append (obj.get_string_member ("expanded_url"));
			    media_builder.append ("\" /></a>");
			    media_list.add (media_builder.str);
			}
        }
        private string parse_entities (Json.Object root, Twitter twitter, ref bool is_mention, ArrayList<string> media_list) {
			bool _is_mention = false;
			var text = root.get_string_member ("text");
			var map = new HashMap<int, Json.Object> ();
			var entities = root.get_object_member ("entities");
			var urls = entities.get_array_member ("urls");
			urls.foreach_element ((array, index_, node) => {
				map_entities (array, index_, node, map);
			});
			var hashtags = entities.get_array_member ("hashtags");
			hashtags.foreach_element ((array, index_, node) => {
				map_entities (array, index_, node, map);
			});
			var user_mentions = entities.get_array_member ("user_mentions");
			user_mentions.foreach_element ((array, index_, node) => {
				map_entities (array, index_, node, map);
				if (node.get_object ().get_string_member ("id_str") == twitter.user_id) {
					_is_mention = true;
				}
			});
			is_mention = _is_mention;
			if (entities.has_member ("media")) {
				var media = entities.get_array_member ("media");
				media.foreach_element ((array, index_, node) => {
					map_entities (array, index_, node, map);
				});
			}
			unichar c = 0;
			int index = 0;
			var builder = new StringBuilder ();
			for (int i = 0; text.get_next_char (ref index, out c); i++) {
				if (map.has_key(i)) {
					var obj = map.get (i);
					process_media (obj, media_list);
					if (obj.has_member ("url")) {
						var url = obj.get_string_member ("url");
						builder.append ("<a href=\"");
						builder.append (url);
						builder.append ("\" title=\"");
						builder.append (obj.get_string_member ("expanded_url"));
						builder.append ("\">");
						builder.append (obj.get_string_member ("display_url"));
						builder.append ("</a>");
						i += (url.length - 1);
						index += (url.length - 1);
					} else if (obj.has_member ("screen_name")) {
						var screen_name = obj.get_string_member ("screen_name");
						builder.append ("@<a href=\"\">");
						builder.append (screen_name);
						builder.append ("</a>");
						i += (screen_name.length);
						index += (screen_name.length);
					} else if (obj.has_member ("text")) {
						var hashtag = obj.get_string_member ("text");
						builder.append ("#<a href=\"\">");
						builder.append (hashtag);
						builder.append ("</a>");
						i += (hashtag.char_count ());
						index += (hashtag.length);
					}
				} else {
					builder.append (c.to_string ());
				}
			}
			text = builder.str;
			return text;
		}
        private void process_tweets (Json.Object root, WebView web_view, Twitter twitter) {
            if (root != null) {
                if (root.has_member ("text")) {
                    string tweet = "";
                    string reply = "";
                    string retweeted = "";
                    var sort_id = root.get_string_member ("id_str");
                    if (root.get_string_member ("in_reply_to_status_id_str") != null) {
                        var i_screen_name = root.get_string_member ("in_reply_to_screen_name");
                        var i_user_id = root.get_string_member ("in_reply_to_user_id_str");
                        var user = twitter.get_user (i_user_id);
                        var i_name = user.get_string_member ("name");
                        var i_profile_image = user.get_string_member ("profile_image_url_https");
                        reply = Trino.Parser.read ("/usr/local/share/trino/html/reply.html");
						reply = (reply.replace ("󠀨id󠀩", sort_id)); //TODO: Does this always work?
						reply = (reply.replace ("󠀨in_reply_to_screen_name󠀩", i_screen_name));
						reply = (reply.replace ("󠀨in_reply_to_name󠀩", i_name));
						reply = (reply.replace ("󠀨in_reply_to_profile_image󠀩", i_profile_image));
						reply = (reply.replace ("󠀨in_reply_to_user_id󠀩", i_user_id));
                    }
                    if (root.has_member ("retweeted_status")) {
                        var r_created_at = root.get_string_member ("created_at");
                        var user = root.get_object_member ("user");
                        var r_name = user.get_string_member ("name");
                        var r_screen_name = user.get_string_member ("screen_name");
                        var r_profile_image = user.get_string_member ("profile_image_url_https");
                        retweeted = Trino.Parser.read ("/usr/local/share/trino/html/retweeted.html");
                        retweeted = (retweeted.replace ("󠀨id󠀩", sort_id));
                        retweeted = (retweeted.replace ("󠀨created_at󠀩", r_created_at));
                        retweeted = (retweeted.replace ("󠀨name󠀩", r_name));
                        retweeted = (retweeted.replace ("󠀨screen_name󠀩", r_screen_name));
                        retweeted = (retweeted.replace ("󠀨profile_image󠀩", r_profile_image));
                        root = root.get_object_member ("retweeted_status");
                    }
                    var media_list = new ArrayList<string> ();
                    bool is_mention = false;
                    var text = parse_entities (root, twitter, ref is_mention, media_list);
                    text = Trino.Parser.escape_html (text);
                    var id = root.get_string_member ("id_str");
                    var created_at = root.get_string_member ("created_at");
                    var user = root.get_object_member ("user");
                    var user_id = user.get_string_member ("id_str");
                    var name = user.get_string_member ("name");
                    var screen_name = user.get_string_member ("screen_name");
                    var is_protected = user.get_boolean_member ("protected");
                    var profile_image = user.get_string_member ("profile_image_url_https");
                    tweet = Trino.Parser.read ("/usr/local/share/trino/html/tweet.html");
                    tweet = (tweet.replace ("󠀨id󠀩", id));
                    tweet = (tweet.replace ("󠀨sort_id󠀩", sort_id));
                    tweet = (tweet.replace ("󠀨text󠀩", text));
                    tweet = (tweet.replace ("󠀨created_at󠀩", created_at));
                    tweet = (tweet.replace ("󠀨name󠀩", name));
                    tweet = (tweet.replace ("󠀨screen_name󠀩", screen_name));
                    tweet = (tweet.replace ("󠀨profile_image󠀩", profile_image));
                    
                    web_view.execute_script ("appendMessage('%s')".printf (tweet));
                    if (twitter.user_id == user_id) {
                        web_view.execute_script ("setSelf('%s')".printf (id));
                    }
                    if (is_protected) {
                        web_view.execute_script ("setProtected('%s')".printf (id));
                    }
                    cached_tweets.set (id, tweet);
                    if (reply != "") {
                        web_view.execute_script ("appendNextMessage('%s')".printf (reply));
                    }
                    if (retweeted != "") {
                        web_view.execute_script ("appendNextMessage('%s')".printf (retweeted));
                    }
                    if (media_list.size > 0) {
                        foreach (string media in media_list) {
                            web_view.execute_script ("appendMedia('%s')".printf (media));
                        }
                    }/*
                    if (root.has_member ("in_reply_to_status_id")) {
                        var 
                        i_tweet = Trino.Parser.read ("/usr/local/share/trino/html/reply.html");
                    }*/
                    
                    var index = accounts.index_of (twitter);
                    var home_view = views.get (index * 2);
                    var mentions_view = views.get (index * 2 + 1);
                    
                    if (is_mention) {
                        if (!twitter.is_stored (id)) {
                            twitter.set_unread (id, true);
                            home_view.execute_script ("addUnread('%s')".printf (id));
                            mentions_view.execute_script ("addUnread('%s')".printf (id));
                            try {
                                var notif = new Notify.Notification(name, text, "twitter");
                                notif.show();
                            } catch (Error e) {
                                error ("Error: %s", e.message);
                            }
                        } else if (twitter.is_unread (id)) {
                            home_view.execute_script ("addUnread('%s')".printf (id));
                            mentions_view.execute_script ("addUnread('%s')".printf (id));
                        }
                        update_launcher ();
                    } else {
                        web_view.execute_script ("hideUnread('%s')".printf (id));
                    }
                } else if (root.has_member ("delete")) {
                    var id = root.get_object_member ("delete")
                                        .get_object_member ("status")
                                        .get_string_member ("id_str");
                    web_view.execute_script ("deleteMessage('%s')".printf (id));
                } else if (root.has_member ("event")) {
                    /*var event = root.get_string_member ("event");
                    var source = root.get_object_member ("source");
                    var target = root.get_object_member ("target");
                    var sid = source.get_string_member ("id_str");
                    var tid = target.get_string_member ("id_str");
                    stdout.printf (event + ": "+ sid + ", " + tid + "\n");
                    web_view.execute_script ("deleteMessage('%s')".printf (id));*/
                }
            }
        }
    }
}

[CCode (array_length = false, array_null_terminated = true)]
private static string[]? importdirs = null;
private static bool tweet = false;

private const GLib.OptionEntry[] options = {
	{ "tweet", 0, 0, OptionArg.NONE, ref tweet, "Show new tweet dialog", null },
	// list terminator
	{ null }
};

public static int main(string[] args) {
	var opt_context = new OptionContext ("- OptionContext example");
	opt_context.set_help_enabled (true);
	opt_context.add_main_entries (options, null);
	try {
		opt_context.parse (ref args);
	} catch (OptionError e) {
		stdout.printf(opt_context.get_help (true, null));
	}
	if (!tweet) {
		Gtk.init(ref args);
		var trino = new Trino.Main();
		return trino.run(args);
	} else {
		stdout.printf ("Hi!\n");
	}
	return 0;
}
