/*
Copyright (C) Arturo Torres Sánchez 2013

This file is part of Trino.

Trino is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Trino is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Trino.  If not, see <http://www.gnu.org/licenses/>.
*/

//valac --pkg libsecret-1 secret.vala
//TODO: Does libsecret come with precise?

namespace Trino {
    public class Keys {
        private Secret.Schema schema;
        public Keys () {
            schema = new Secret.Schema ("org.pantheon.trino", Secret.SchemaFlags.NONE,
                                            "user_id", Secret.SchemaAttributeType.STRING,
                                            "type", Secret.SchemaAttributeType.STRING);
        }
        public void add (string access_token, string access_token_secret) {
            try {
                var user_id = access_token.split ("-")[0];
                Secret.password_store_sync (schema, Secret.COLLECTION_DEFAULT,
                                            "Twitter credentials", access_token, null,
                                            "user_id", user_id,
                                            "type", "access_token");
                Secret.password_store_sync (schema, Secret.COLLECTION_DEFAULT,
                                            "Twitter credentials", access_token_secret, null,
                                            "user_id", user_id,
                                            "type", "access_token_secret");
            } catch (Error e) {
                message ("Error: %s", e.message);
            }
        }
        public string get_access_token (string user_id) {
            string res = "";
            try {
                res = Secret.password_lookup_sync (schema, null,
                                                   "user_id", user_id,
                                                   "type", "access_token");
            } catch (Error e) {
                message ("Error: %s", e.message);
            }
            return res;
        }
        public string get_access_token_secret (string user_id) {
            string res = "";
            try {
                res = Secret.password_lookup_sync (schema, null,
                                                   "user_id", user_id,
                                                   "type", "access_token_secret");
            } catch (Error e) {
                message ("Error: %s", e.message);
            }
            return res;
        }
    }
}
