/*
Copyright (C) Arturo Torres Sánchez 2013

This file is part of Trino.

Trino is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Trino is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Trino.  If not, see <http://www.gnu.org/licenses/>.
*/

//valac --pkg gtk+-3.0 --pkg granite --pkg libnotify --pkg appindicator3-0.1 --pkg libsoup-2.4 --pkg gee-1.0 --pkg json-glib-1.0 --pkg libsecret-1 --pkg gio-2.0 --target-glib=2.32 --pkg webkitgtk-3.0 --thread twitter.vala oauth.vala secret.vala

using Gee;
using Granite.Services;
using Soup;
using Sqlite;
using OAuth;
using WebKit;

namespace Trino {
    public class Twitter : GLib.Object {
        private string consumer_key;
        private string consumer_secret;
        private string home_lastid;
        private string mentions_lastid;
        private int refresh_time;
        private Database db;
        private Client client;
        private Session session;
        private GLib.Settings settings;
        private Keys keys;
        public string user_id { get; private set; }
        public string screen_name { get; private set; }
        public int characters_reserved_per_media { get; private set; }
        public int short_url_length { get; private set; }
        public int short_url_length_https { get; private set; }
        public Twitter (string user_id) {
            settings = new GLib.Settings ("org.pantheon.trino");
            refresh_time = settings.get_int ("refresh-time") * 60;
            
            consumer_key = "JQp0519PS6NJRt59WRWryg";
            consumer_secret = "jYXHAU2tqWBaeg79JhKhjWdGu1NK2XMODxWabZWt1c";
            keys = new Keys ();
            session = new SessionAsync ();
            client = new Client ("", consumer_key, new HMAC_SHA1(consumer_secret));
            client.set_token(keys.get_access_token(user_id), keys.get_access_token_secret(user_id));
            this.user_id = user_id;
            this.screen_name = _get_screen_name ();
            get_configs ();
            
            //Paths.initialize ("trino", "");
            //var paths = new Paths();
            //var folder = paths.user_data_folder;
            Database.open ("/home/r2d221/.local/share/trino/%s.db".printf (user_id), out db);
            db.exec ("CREATE TABLE unread (id TEXT, unread BOOLEAN)");
        }
        private Message api_request (string method, string url, ...) {
            HashMultiMap<string, string> args = new HashMultiMap<string, string>();
            string args_str = "?";
            string escape_chars = " !\"#$%&'()*+,/:;<=>?@[\\]^`{|}";
            var l = va_list();
            while (true) {
                string? key = l.arg();
                if (key == null) {
                    break;
                }
                string val = l.arg();
                args.set (key, val);
                args_str = args_str.concat (URI.encode (key, escape_chars), "=", URI.encode (val, escape_chars), "&");
            }
            args_str = args_str.slice (0, args_str.length - 1);
            
            var message = new Message(method, url + args_str);
            var oauth_headers = client.authenticate(method, url, args);
            message.request_headers.append("Authorization", oauth_headers["Authorization"]);
            return message;
        }
        public static string get_access_token (WebView view) {
            var settings = new GLib.Settings ("org.pantheon.trino");
            var keys = new Keys ();
            var consumer_key = "JQp0519PS6NJRt59WRWryg";
            var consumer_secret = "jYXHAU2tqWBaeg79JhKhjWdGu1NK2XMODxWabZWt1c";
            var session = new SessionAsync ();
            var client = new Client ("", consumer_key, new HMAC_SHA1(consumer_secret));
            
            string request_auth = client.request_token("POST", "http://api.twitter.com/oauth/request_token", "trino://callback")["Authorization"];
            var message = new Message("POST", "http://api.twitter.com/oauth/request_token");
            message.request_headers.append("Authorization", request_auth);
            session.send_message(message);
            var response = (string) message.response_body.data;
            var response_map = Trino.Parser.parse_response (response);
            //print (response + "\n");
            var oauth_token = response_map.get ("oauth_token");
            var oauth_token_secret = response_map.get ("oauth_token_secret");
            var access_token = "";
            try {
                string url_to_open_in_browser = client.auth_token("https://api.twitter.com/oauth/authorize?oauth_token=%s", oauth_token, oauth_token_secret, "true");
                view.load_uri (url_to_open_in_browser);
                view.navigation_policy_decision_requested.connect ((frame, request, action, decision) => {
                    var callback = request.get_uri();
                    if (callback.has_prefix ("trino://callback?")) {
                        decision.ignore ();
                        callback = callback.replace ("trino://callback?", "");
                        var callback_map = Trino.Parser.parse_response (callback);
                        oauth_token = callback_map.get ("oauth_token");
                        var oauth_verifier = callback_map.get ("oauth_verifier");
                        string access_auth = client.access_token("POST", "https://api.twitter.com/oauth/access_token", oauth_token, oauth_verifier)["Authorization"];
                        message = new Message("POST", "http://api.twitter.com/oauth/access_token");
                        message.request_headers.append("Authorization", access_auth);
                        session.send_message(message);
                        response = (string) message.response_body.data;
                        response_map = Trino.Parser.parse_response (response);
                        access_token = response_map.get ("oauth_token");
                        var access_token_secret = response_map.get ("oauth_token_secret");
                        keys.add (access_token, access_token_secret);
                        string[] users = settings.get_strv ("users");
                        users += access_token.split ("-")[0];
                        settings.set_strv ("users", users);
                        settings.sync();
                        return true;
                    }
                    return false;
                });
            } catch (OAuth.Error e) {
                stdout.printf ("OAuth.Error: Invalid argument");
            }
            return access_token;
        }
        private void get_configs () {
            var message = api_request ("GET", "https://api.twitter.com/1.1/help/configuration.json");
            session.send_message(message);
            var data = (string) message.response_body.data;
            var parser = new Json.Parser ();
            try {
                parser.load_from_data (data);
                var object = parser.get_root ().get_object ();
                characters_reserved_per_media = (int) object.get_int_member ("characters_reserved_per_media");
                short_url_length = (int) object.get_int_member ("short_url_length");
                short_url_length_https = (int) object.get_int_member ("short_url_length_https");
            } catch (GLib.Error e) {
                stdout.printf ("Error: %s", e.message);
            }
        }
        private string _get_screen_name () {
            var message = api_request ("GET", "https://api.twitter.com/1.1/account/settings.json");
            session.send_message(message);
            var data = (string) message.response_body.data;
            var parser = new Json.Parser ();
            string screen_name = "";
            try {
                parser.load_from_data (data);
                var object = parser.get_root ().get_object ();
                screen_name = object.get_string_member ("screen_name");
            } catch (GLib.Error e) {
                stdout.printf ("Error: %s", e.message);
            }
            return screen_name;
        }
        public uint send_tweet (string tweet, string? reply_id) {
            Message message;
            if (reply_id == null) {
                message = api_request ("POST", "https://api.twitter.com/1.1/statuses/update.json", "status", tweet);
            } else {
                message = api_request ("POST", "https://api.twitter.com/1.1/statuses/update.json", "status", tweet, "in_reply_to_status_id", reply_id);
            }
            session.send_message(message);
            return message.status_code;
        }
        /*I don't know what the fuck I did here, but it works*/
        public uint send_image (string tweet, string? reply_id, string uri) {
            Multipart message_parts = new Multipart("multipart/form-data");
            message_parts.append_form_string("status", tweet);
            if (reply_id != null) {
                message_parts.append_form_string("in_reply_to_status_id", reply_id);
            }
            var file = File.new_for_path (uri);
            var fileinfo = file.query_info ("*", FileQueryInfoFlags.NONE);
            var content_type = fileinfo.get_content_type ();
            
            uint8[] contents;
            FileUtils.get_data (uri, out contents);
            
            Buffer buffer = new Buffer (MemoryUse.COPY, contents);
            message_parts.append_form_file("media[]", uri, content_type, buffer);
            
            Message message = Form.request_new_from_multipart("https://api.twitter.com/1.1/statuses/update_with_media.json", message_parts);
            
            HashMultiMap<string, string> args = new HashMultiMap<string, string>();
            var oauth_headers = client.authenticate("POST", "https://api.twitter.com/1.1/statuses/update_with_media.json", args);
            message.request_headers.append("Authorization", oauth_headers["Authorization"]);
            
            session.send_message(message);
            return message.status_code;
        }
        public uint delete_tweet (string id) {
            Message message;
            message = api_request ("POST", "https://api.twitter.com/1.1/statuses/destroy/" + id + ".json");
            session.send_message(message);
            return message.status_code;
        }
        public uint retweet (string retweet_id) {
            Message message;
            message = api_request ("POST", "https://api.twitter.com/1.1/statuses/retweet/" + retweet_id + ".json");
            session.send_message(message);
            return message.status_code;
        }
        public uint favorite (string favorite_id) {
            Message message;
            message = api_request ("POST", "https://api.twitter.com/1.1/favorites/create.json", "id", favorite_id);
            session.send_message(message);
            return message.status_code;
        }
        public bool is_stored (string id) {
            bool exists = false;
            db.exec ("SELECT * FROM unread WHERE id='%s'".printf (id), (n_columns, values, column_names) => {
                exists = true;
                return 0;
            });
            return exists;
        }
        public void set_unread (string id, bool unread) {
            bool exists = is_stored (id);
            if (exists) {
                string errmsg;
                db.exec ("UPDATE unread SET unread='%s' WHERE id='%s'".printf (unread.to_string(), id), null, out errmsg);
            } else {
                string errmsg;
                db.exec ("INSERT INTO unread VALUES ('%s','true')".printf (id), null, out errmsg);
            }
        }
        public bool is_unread (string id) {
            bool unread = true;
            db.exec ("SELECT * FROM unread WHERE id='%s'".printf (id), (n_columns, values, column_names) => {
                unread = bool.parse(values[1]);
                return 0;
            });
            return unread;
        }
        public int get_unread_count () {
            int unread = 0;
            db.exec ("SELECT * FROM unread WHERE unread='true'", (n_columns, values, column_names) => {
                unread++;
                return 0;
            });
            return unread;
        }
        public Json.Object? get_user (string user_id) {
            Message message;
            message = api_request ("GET", "https://api.twitter.com/1.1/users/show.json", "user_id", user_id);
            session.send_message(message);
            var data = (string) message.response_body.data;
            var parser = new Json.Parser ();
            try {
                parser.load_from_data (data);
                var user = parser.get_root ().get_object ();
                return user;
            } catch (GLib.Error e) {
                return null;
            }
        }
        public Json.Object? get_tweet (string id) {
            Message message;
            message = api_request ("GET", "https://api.twitter.com/1.1/statuses/show.json", "id", id, "include_entities", "true");
            session.send_message(message);
            var data = (string) message.response_body.data;
            stdout.printf (data + "\n");
            var parser = new Json.Parser ();
            try {
                parser.load_from_data (data);
                var tweet = parser.get_root ().get_object ();
                return tweet;
            } catch (GLib.Error e) {
                return null;
            }
        }
        public delegate void DelegateType(Json.Object json, WebView view, Twitter twitter);
        public void home_timeline (DelegateType d, WebView view) {
            /* STREAMING API */
            home_stream (d, view);
            /* REST API */
            home_rest (d, view);
            Timeout.add_seconds_full (Priority.DEFAULT, refresh_time, () => {
                home_rest(d, view);
                return true;
            });
        }
        private void home_stream (DelegateType d, WebView view) {
            var message = api_request ("GET", "https://userstream.twitter.com/1.1/user.json");
            message.got_chunk.connect ((chunk) => {
                var parser = new Json.Parser ();
                var data = ((string) chunk.data).split ("\n")[0];
                try {
                    parser.load_from_data (data);
                    var root = parser.get_root ();
                    if (root != null) {
                        var obj = root.get_object ();
                        if (obj.has_member ("event")) {
                            var event = obj.get_string_member ("event");
                            if (event == "user_update") {
                                var target = obj.get_object_member ("target");
                                var screen_name = target.get_string_member ("screen_name");
                                this.screen_name = screen_name;
                            }
                        }
                        d(obj, view, this);
                    }
                } catch (GLib.Error e) {
                    stdout.printf ("Error: %s", e.message);
                }
            });
            session.queue_message(message, null);
        }
        private void home_rest (DelegateType d, WebView view) {
            Message message;
            if (home_lastid == null) {
                message = api_request ("GET", "https://api.twitter.com/1.1/statuses/home_timeline.json", "include_entities", "true");
            } else {
                message = api_request ("GET", "https://api.twitter.com/1.1/statuses/home_timeline.json", "include_entities", "true", "since_id", home_lastid);
            }
            session.send_message(message);
            var data = (string) message.response_body.data;
            //stdout.printf (data + "\n");
            var parser = new Json.Parser ();
            try {
                parser.load_from_data (data);
                var array = parser.get_root ().get_array ();
                if (array.get_length () > 0) {
                    var last = array.get_object_element (0);
                    home_lastid = last.get_string_member ("id_str");
                }
                array.foreach_element ((array, index_, element_node) => {
                    d(element_node.get_object (), view, this);
                });
            } catch (GLib.Error e) {
                stdout.printf ("Error: %s", e.message);
            }
        }
        public void mentions_timeline (DelegateType d, WebView view) {
            /* STREAMING API */
            mentions_stream (d, view);
            /* REST API */
            mentions_rest (d, view);
            Timeout.add_seconds_full (Priority.DEFAULT, refresh_time, () => {
                mentions_rest(d, view);
                return true;
            });
        }
        private void mentions_stream (DelegateType d, WebView view) {
            var message = api_request ("POST", "https://stream.twitter.com/1.1/statuses/filter.json", "track", "@%s".printf (screen_name));
            message.got_chunk.connect ((chunk) => {
                var parser = new Json.Parser ();
                var data = ((string) chunk.data).split ("\n")[0];
                try {
                    parser.load_from_data (data);
                    var root = parser.get_root ();
                    if (root != null) {
                        var obj = root.get_object ();
                        d(obj, view, this);
                    }
                } catch (GLib.Error e) {
                    stdout.printf ("Error: %s", e.message);
                }
            });
            session.queue_message(message, null);
        }
        private void mentions_rest (DelegateType d, WebView view) {
            Message message;
            if (mentions_lastid == null) {
                message = api_request ("GET", "https://api.twitter.com/1.1/statuses/mentions_timeline.json", "include_entities", "true");
            } else {
                message = api_request ("GET", "https://api.twitter.com/1.1/statuses/mentions_timeline.json", "include_entities", "true", "since_id", mentions_lastid);
            }
            session.send_message(message);
            var data = (string) message.response_body.data;
            var parser = new Json.Parser ();
            try {
                parser.load_from_data (data);
                var array = parser.get_root ().get_array ();
                if (array.get_length () > 0) {
                    var last = array.get_object_element (0);
                    mentions_lastid = last.get_string_member ("id_str");
                }
                array.foreach_element ((array, index_, element_node) => {
                    d(element_node.get_object (), view, this);
                });
            } catch (GLib.Error e) {
                stdout.printf ("Error: %s", e.message);
            }
        }
    }
}
